import logging
from abc import ABC, abstractmethod
from typing import Dict, Optional, List

from inds.core.consumer import Consumer
from inds.core.decorators import action
from inds.core.producer import Producer


class Node(ABC):
    def __init__(self, name: str) -> None:
        self.logger = logging.getLogger(__name__)

        self.name = name
        self.consumers: Dict[str, Consumer] = dict()
        self.producers: Dict[str, Producer] = dict()
        self.running: bool = False

        self.actions: Dict[str, callable] = dict()
        for name in dir(self.__class__):
            func = getattr(self.__class__, name)
            if isinstance(func, action):
                self.actions[name] = func

    def add_consumer(self, consumer: Consumer):
        self.consumers[consumer.name] = consumer

    def add_producer(self, producer: Producer):
        self.producers[producer.name] = producer

    def start(self):
        self.running = True
        self.run()

    def perform_action(self, action_name: str, args: List, kwargs: Dict):
        try:
            func = getattr(self, action_name)
        except AttributeError:
            return

        if func is not None:
            func(self, *args, **kwargs)

    @staticmethod
    def get_config_entries() -> Dict:
        """Return the entries of the configuration of this kind of nodes."""
        return {}

    @staticmethod
    @abstractmethod
    def get_config_categories() -> List[str]:
        """Return the categories of the configuration of this kind of nodes."""

    @staticmethod
    @abstractmethod
    def get_config_key() -> str:
        """Return the key of the configuration of this kind of nodes."""

    @staticmethod
    @abstractmethod
    def build(name: str, config: Dict) -> Optional['Node']:
        """Create an instance of the node from a configuration."""

    @abstractmethod
    def run(self):
        """Run the processing loop."""

    def stop(self):
        """Stop the node."""
        self.running = False

    def disconnect(self):
        """Remove all connections to other nodes (producers and consumers)."""
        for consumer in self.consumers.values():
            consumer.unregister_all()

        for producer in self.producers.values():
            producer.unregister_all()
